from threading import RLock
from typing import Any, Callable, Dict, Tuple


class RunOnce(object):
    """Run `func` once, the uniqueness is defined by `key_func`

    :param func: the function to run only once with this wrapper instance
    :param key_func: the unique key determined by arguments of `func`
    :param lock_type: lock class type for thread safe

    :Examples:
    >>> r = RunOnce(max, lambda *args, **kwargs: id(args[0]))
    >>> a1 = [0, 1]
    >>> a2 = [0, 1]
    >>> assert 1 == r(a1) # will trigger max
    >>> assert 1 == r(a1) # will get the result from cache
    >>> assert 1 == r(a2) # will trigger max again because id(a1)!=id(a2)

    :Notice:
    * Hash collision is the concern of the user, not this class, your
      `key_func` should avoid any potential collision
    * `func` can have no return
    * For concurrent calls of this wrapper, only one will trigger `func` other
      calls will be blocked until the first call returns an result
    * This class is cloudpicklable, but unpickled instance does NOT share the same
      context with the original one
    """

    def __init__(self, func: Callable, key_func: Callable, lock_type: type = RLock):
        self._func = func
        self._key_func = key_func
        self._lock_type = lock_type
        self._init_locks()

    def __getstate__(self) -> Dict[str, Any]:
        d = dict(self.__dict__)
        del d["_lock"]
        del d["_locks"]
        del d["_store"]
        return d

    def __setstate__(self, members: Any) -> None:
        self.__dict__.update(members)
        self._init_locks()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        key = self._key_func(*args, **kwargs)
        lock = self._get_lock(key)
        with lock:
            found, res = self._try_get(key)
            if found:
                return res
            res = self._func(*args, **kwargs)
            self._update(key, res)
            return res

    def _get_lock(self, key) -> Any:
        with self._lock:
            if key not in self._locks:
                self._locks[key] = self._lock_type()
            return self._locks[key]

    def _try_get(self, key) -> Tuple[bool, Any]:
        with self._lock:
            if key in self._store:
                return True, self._store[key]
            return False, None

    def _update(self, key, value) -> None:
        with self._lock:
            self._store[key] = value

    def _init_locks(self):
        self._lock = self._lock_type()
        self._locks: Dict[int, Any] = {}
        self._store: Dict[int, Any] = {}



from typing import Union, Any
from triad.exceptions import NoneArgumentError


def assert_or_throw(
    bool_exp: bool, exception: Union[None, str, Exception] = None
) -> None:
    """Assert on expression and throw custom exception

    :param bool_exp: boolean expression to assert on
    :param exception: a custom Exception instance, or any other object that
        will be stringfied and instantiate an AssertionError
    """
    if not bool_exp:
        if isinstance(exception, Exception):
            raise exception
        if isinstance(exception, str):
            raise AssertionError(exception)
        if exception is None:
            raise AssertionError()
        raise AssertionError(str(exception))


def assert_arg_not_none(obj: Any, arg_name: str = "", msg: str = "") -> None:
    """Assert an argument is not None, otherwise raise exception

    :param obj: argument value
    :param arg_name: argument name, if None or empty, it will use `msg`
    :param msg: only when `arg_name` is None or empty, this value is used

    :raises NoneArgumentError: with `arg_name` or `msg`
    """
    if obj is None:
        if arg_name != "" and arg_name is not None:
            msg = f"{arg_name} can't be None"
        msg = msg or ""
        raise NoneArgumentError(msg)



from threading import RLock
from typing import Dict, Tuple
from urllib.parse import urlparse

from fs import open_fs, tempfs, memoryfs
from fs.base import FS as FSBase
from fs.mountfs import MountFS
from triad.utils.hash import to_uuid
import os


class FileSystem(MountFS):
    """A unified filesystem based on PyFileSystem2. The special requirement
    for this class is that all paths must be absolute path with scheme.
    To customize different file systems, you should override `create_fs`
    to provide your own configured file systems.

    :Examples:
    >>> fs = FileSystem()
    >>> fs.writetext("mem://from/a.txt", "hello")
    >>> fs.copy("mem://from/a.txt", "mem://to/a.txt")

    :Notice:
    If a path is not a local path, it must include the scheme and `netloc`
    (the first element after `://`)

    :param auto_close: If `True` (the default), the child filesystems
      will be closed when `MountFS` is closed.
    """

    def __init__(self, auto_close: bool = True):
        super().__init__(auto_close)
        self._fs_store: Dict[str, FSBase] = {}
        self._in_create = False
        self._fs_lock = RLock()

    def create_fs(self, root: str) -> FSBase:
        """create a PyFileSystem instance from `root`. `root` is in the
        format of `/` if local path, else `<scheme>://<netloc>`.

        You should override this method to provide custom instances, for
        example, if you want to create an S3FS with certain parameters.

        :param root: `/` if local path, else `<scheme>://<netloc>`
        """
        if root.startswith("temp://"):
            fs = tempfs.TempFS(root[len("temp://") :])
            return fs
        if root.startswith("mem://"):
            fs = memoryfs.MemoryFS()
            return fs
        return open_fs(root)

    def _delegate(self, path) -> Tuple[FSBase, str]:
        with self._fs_lock:
            if self._in_create:  # pragma: no cover
                return super()._delegate(path)
            self._in_create = True
            fp = _FSPath(path)
            if fp.root not in self._fs_store:
                self._fs_store[fp.root] = self.create_fs(fp.root)
                self.mount(to_uuid(fp.root), self._fs_store[fp.root])
            self._in_create = False
        m_path = os.path.join(to_uuid(fp.root), fp.relative_path)
        return super()._delegate(m_path)


class _FSPath(object):
    def __init__(self, path: str):
        if path is None:
            raise ValueError("path can't be None")
        path = self._modify_path(path)
        if path.startswith("file://"):
            path = path[6:]
        if path.startswith("/"):
            self._scheme = ""
            self._root = "/"
            self._path = os.path.abspath(path)
        else:
            uri = urlparse(path)
            if uri.scheme == "" and not path.startswith("/"):
                raise ValueError(
                    f"invalid {path}, must be abs path either local or with scheme"
                )
            self._scheme = uri.scheme
            if uri.netloc == "":
                raise ValueError(f"invalid path {path}")
            self._root = uri.scheme + "://" + uri.netloc
            self._path = uri.path
        self._path = self._path.lstrip("/")
        # if self._path == "":
        #    raise ValueError(f"invalid path {path}")

    @property
    def scheme(self) -> str:
        return self._scheme

    @property
    def root(self) -> str:
        return self._root

    @property
    def relative_path(self) -> str:
        return self._path

    def _modify_path(self, path: str) -> str:
        """to fix things like /s3:/a/b.txt -> s3://a/b.txt
        """
        if path.startswith("/"):
            p = path.find("/", 1)
            if p > 1 and path[p - 1] == ":":
                scheme = path[1 : p - 1]
                return scheme + "://" + path[p + 1 :]
        return path



import datetime
import importlib
import inspect
from pydoc import locate
from typing import Any, Dict, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd
import six
from ciso8601 import parse_datetime

EMPTY_ARGS: List[Any] = []
EMPTY_KWARGS: Dict[str, Any] = {}


def str_to_type(
    s: str, expected_base_type: Optional[type] = None, first: bool = False
) -> type:
    """Given a string expression, find the first/last type from all import libraries.
    If the expression contains `.`, it's supposed to be a full path of the type
    including modules, and this function will use `pydoc.locate` to find the type.
    Otherwise, it will firstly check if it is a built in type, then search
    on the call stack.

    :param s: type expression, for example `triad.utils.iter.Slicer` or `str`
    :param expected_base_type: base class type that must satisfy, defaults to None
    :param first: to return the nearest (True) or farthest (False) matching one

    :raises TypeError: unable to find a matching type

    :return: found type
    """
    t = list(str_to_types(s, expected_base_type))
    if len(t) == 0:
        if expected_base_type is not None:
            raise TypeError(f"{s} is not type {str(expected_base_type)}")
        else:
            raise TypeError(f"{s} is not type")
    if first:
        return t[0]
    return t[-1]


def str_to_types(s: str, expected_base_type: Optional[type] = None) -> Iterable[type]:
    """Given a string expression, find the type from all import libraries.
    If the expression contains `.`, it's supposed to be a full path of the type
    including modules, and this function will use `pydoc.locate` to find the type.
    Otherwise, it will firstly check if it is a built in type, then search
    on the call stack, starting from nearest level.

    :param s: type expression, for example `triad.utils.iter.Slicer` or `str`
    :param expected_base_type: base class type that must satisfy, defaults to None

    :yield: found types, starting from the top on call stack
    """
    if "." in s:
        t = locate(s)
        if isinstance(t, type) and (
            expected_base_type is None or issubclass(t, expected_base_type)
        ):
            yield t
    else:
        for tt in str_to_types("builtins." + s, expected_base_type):
            yield tt
        for frm in inspect.stack():
            t = frm[0].f_globals["__name__"] + "." + s
            for tt in str_to_types(str(t), expected_base_type):
                yield tt


def str_to_instance(
    s: str,
    expected_base_type: Optional[type] = None,
    first: bool = False,
    args: List[Any] = EMPTY_ARGS,
    kwargs: Dict[str, Any] = EMPTY_KWARGS,
) -> Any:
    """Use :func:'~triad.utils.convert.str_to_type' to find a matching type
    and instantiate

    :param s: see :func:'~triad.utils.convert.str_to_type'
    :param expected_base_type: see :func:'~triad.utils.convert.str_to_type'
    :param first: see :func:'~triad.utils.convert.str_to_type'
    :param args: args to instantiate the type
    :param kwargs: kwargs to instantiate the type

    :return: the instantiated the object
    """
    t = str_to_type(s, expected_base_type, first)
    return t(*args, **kwargs)


def to_type(
    s: Any, expected_base_type: Optional[type] = None, first: bool = False
) -> type:
    """Convert an object `s` to `type`
    * if `s` is `str`: see :func:'~triad.utils.convert.str_to_type'
    * if `s` is `type`: check `expected_base_type` and return itself
    * else: check `expected_base_type` and return itself

    :param s: see :func:'~triad.utils.convert.str_to_type'
    :param expected_base_type: see :func:'~triad.utils.convert.str_to_type'
    :param first: see :func:'~triad.utils.convert.str_to_type'

    :raises TypeError: if no matching type found

    :return: the matching type
    """
    if isinstance(s, str):
        return str_to_type(s, expected_base_type, first)
    if isinstance(s, type):
        if expected_base_type is None or issubclass(s, expected_base_type):
            return s
        raise TypeError(f"Type mismatch {s} expected {expected_base_type}")
    t = type(s)
    if expected_base_type is None or issubclass(t, expected_base_type):
        return t
    raise TypeError(f"Type mismatch {s} expected {expected_base_type}")


def to_instance(
    s: Any,
    expected_base_type: Optional[type] = None,
    first: bool = False,
    args: List[Any] = EMPTY_ARGS,
    kwargs: Dict[str, Any] = EMPTY_KWARGS,
) -> Any:
    """If s is str or type, then use :func:'~triad.utils.convert.to_type' to find
    matching type and instantiate. Otherwise return s if it matches constraints

    :param s: see :func:'~triad.utils.convert.to_type'
    :param expected_base_type: see :func:'~triad.utils.convert.to_type'
    :param first: see :func:'~triad.utils.convert.to_type'
    :param args: args to instantiate the type
    :param kwargs: kwargs to instantiate the type

    :raises ValueError: if s is an instance but not a (sub)type of `expected_base_type`
    :raises TypeError: if s is an instance, args and kwargs must be empty

    :return: the instantiated object
    """
    if s is None:
        raise ValueError("None can't be converted to instance")
    if isinstance(s, (str, type)):
        t = to_type(s, expected_base_type, first)
        return t(*args, **kwargs)
    else:
        if expected_base_type is not None and not isinstance(s, expected_base_type):
            raise TypeError(f"{str(s)} is not a subclass of {str(expected_base_type)}")
        if len(args) > 0 or len(kwargs) > 0:
            raise ValueError(f"Can't instantiate {str(s)} with different parameters")
        return s


def to_function(s: Any, first: bool = False) -> Any:  # noqa: C901
    """For an expression, it will try to find the nearest or farthest
    correspondent function in all imported libraries.
    Otherwise if it is a callable and not a class, it will return the input itself.

    :params s: a string expression or a callable
    :param first: to return the nearest (True) or farthest (False) matching one

    :raises AttributeError: if unable to find such a function

    :return: the nearest (first=True) or farthest (False) matching function
    """
    funcs = list(to_functions(s))
    if len(funcs) == 0:
        raise AttributeError(f"{s} is not a valid function")
    if first:
        return funcs[0]
    return funcs[-1]


def to_functions(s: Any) -> Iterable[Any]:  # noqa: C901
    """For an expression, it will try to find all correspondent function
    in all imported libraries, starting from the nearest on call stack.
    Otherwise if it is a callable and not a class, it will return the input itself.

    :params s: a string expression or a callable

    :raises AttributeError: if unable to find such a function

    :yield: matching functions
    """
    if s is None:
        raise AttributeError("None can't be converted to function")
    func: Any = None
    if isinstance(s, str):
        if "." in s:
            mod_name, func_name = s.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            try:
                func = getattr(mod, func_name)
            except AttributeError:
                func = None
        else:
            for f in to_functions("builtins." + s):
                yield f
            for frm in inspect.stack():
                t = frm[0].f_globals["__name__"] + "." + s
                for f in to_functions(str(t)):
                    yield f
    else:
        func = s
    if callable(func) and not isinstance(func, six.class_types):
        yield func


def get_full_type_path(obj: Any) -> str:
    """Get the full module path of the type (if `obj` is class or function) or type
    of the instance (if `obj` is an object instance)

    :param obj: a class/function type or an object instance
    :raises TypeError: if `obj` is None, lambda, or neither a class or a function
    :return: full path string
    """
    if obj is not None:
        if inspect.isclass(obj):
            return "{}.{}".format(obj.__module__, obj.__name__)
        if inspect.isfunction(obj):
            if obj.__name__.startswith("<lambda"):
                raise TypeError("Can't get full path for lambda functions")
            return "{}.{}".format(obj.__module__, obj.__name__)
        if isinstance(obj, object):
            return "{}.{}".format(obj.__class__.__module__, obj.__class__.__name__)
    raise TypeError(f"Unable to get type full path from {obj}")


def to_bool(obj: Any) -> bool:
    """Convert an object to python bool value. It can handle values
    like `True`, `true`, `yes`, `1`, etc

    :param obj: object
    :raises TypeError: if failed to convert
    :return: bool value
    """
    if obj is None:
        raise TypeError("None can't convert to bool")
    o = str(obj).lower()
    if o in ["true", "yes", "1"]:
        return True
    if o in ["false", "no", "0"]:
        return False
    raise TypeError(f"{o} can't convert to bool")


def to_datetime(obj: Any) -> datetime.datetime:
    """Convert an object to python datetime. If the object is a
    string, it will use `ciso8601.parse_datetime` to parse

    :param obj: object
    :raises TypeError: if failed to convert
    :return: datetime value
    """
    if obj is None:
        raise TypeError("None can't convert to datetime")
    if isinstance(obj, datetime.datetime):
        return obj
    if isinstance(obj, datetime.date):
        return datetime.datetime(obj.year, obj.month, obj.day)
    if isinstance(obj, str):
        try:
            return parse_datetime(obj)
        except Exception as e:
            raise TypeError(f"{obj} can't convert to datetime", e)
    raise TypeError(f"{type(obj)} {obj} can't convert to datetime")


def to_timedelta(obj: Any) -> datetime.timedelta:
    """Convert an object to python datetime.

    If the object is a string, `min` or `-inf` will return `timedelta.min`,
    `max` or `inf` will return `timedelta.max`; if the object is a number,
    the number will be used as the seconds argument; Otherwise it will use
    `pandas.to_timedelta` to parse the object.

    :param obj: object
    :raises TypeError: if failed to convert
    :return: timedelta value
    """
    if obj is None:
        raise TypeError("None can't convert to timedelta")
    if isinstance(obj, datetime.timedelta):
        return obj
    if np.isreal(obj):
        return datetime.timedelta(seconds=float(obj))
    try:
        return pd.to_timedelta(obj).to_pytimedelta()
    except Exception as e:
        if isinstance(obj, str):
            obj = obj.lower()
            if obj in ["min", "-inf"]:
                return datetime.timedelta.min
            elif obj in ["max", "inf"]:
                return datetime.timedelta.max
        raise TypeError(f"{type(obj)} {obj} can't convert to timedelta", e)


def as_type(obj: Any, target: type) -> Any:
    """Convert `obj` into `target` type

    :param obj: input object
    :param target: target type

    :return: object in the target type
    """
    if issubclass(type(obj), target):
        return obj
    if target == bool:
        return to_bool(obj)
    if target == datetime.datetime:
        return to_datetime(obj)
    if target == datetime.timedelta:
        return to_timedelta(obj)
    return target(obj)


def to_size(exp: Any) -> int:
    """Convert input value or expression to size
    For expression string, it must be in the format of
    `<value>` or `<value><unit>`. Value must be 0 or positive,
    default unit is byte if not provided. Unit can be `b`, `byte`,
    `k`, `kb`, `m`, `mb`, `g`, `gb`, `t`, `tb`.

    Args:
        exp (Any): expression string or numerical value

    Raises:
        ValueError: for invalid expression
        ValueError: for negative values

    Returns:
        int: size in byte
    """
    n, u = _parse_value_and_unit(exp)
    assert n >= 0.0, "Size can't be negative"
    if u in ["", "b", "byte", "bytes"]:
        return int(n)
    if u in ["k", "kb"]:
        return int(n * 1024)
    if u in ["m", "mb"]:
        return int(n * 1024 * 1024)
    if u in ["g", "gb"]:
        return int(n * 1024 * 1024 * 1024)
    if u in ["t", "tb"]:
        return int(n * 1024 * 1024 * 1024 * 1024)
    raise ValueError(f"Invalid size expression {exp}")


def _parse_value_and_unit(exp: Any) -> Tuple[float, str]:
    try:
        assert exp is not None
        if isinstance(exp, (int, float)):
            return float(exp), ""
        exp = str(exp).replace(" ", "").lower()
        i = 1 if exp.startswith("-") else 0
        while i < len(exp):
            if (exp[i] < "0" or exp[i] > "9") and exp[i] != ".":
                break
            i += 1
        return float(exp[:i]), exp[i:]
    except (ValueError, AssertionError):
        raise ValueError(f"Invalid expression {exp}")



from triad.utils.string import assert_triad_var_name
from pytest import raises


def test_assert_triad_var_name():
    raises(AssertionError, lambda: assert_triad_var_name(None))
    raises(AssertionError, lambda: assert_triad_var_name(""))
    raises(AssertionError, lambda: assert_triad_var_name(" "))
    raises(AssertionError, lambda: assert_triad_var_name("1"))
    raises(AssertionError, lambda: assert_triad_var_name(123))
    raises(AssertionError, lambda: assert_triad_var_name("_"))
    raises(AssertionError, lambda: assert_triad_var_name("__"))
    raises(AssertionError, lambda: assert_triad_var_name("a "))
    raises(AssertionError, lambda: assert_triad_var_name("a a"))
    assert "a" == assert_triad_var_name("a")
    assert "_1" == assert_triad_var_name("_1")
    assert "_a_1" == assert_triad_var_name("_a_1")



from triad.utils.threading import RunOnce
import concurrent
from triad.utils.assertion import assert_or_throw
import cloudpickle


def test_runonce():
    def con(df, arr, n=1):
        arr[0] += n
        return df

    arr = [0]
    sc = RunOnce(con, lambda *args, **kwargs: id(args[0]))
    sc = cloudpickle.loads(cloudpickle.dumps(sc))
    df1 = [[0]]
    df2 = [[0]]
    assert sc(df1, arr) is df1
    assert 1 == arr[0]
    assert sc(df1, arr) is df1
    assert 1 == arr[0]
    assert sc(df2, arr, n=2) is df2
    assert 3 == arr[0]
    assert sc(df2, arr, n=3) is df2
    assert 3 == arr[0]
    arr[0] = 0
    # multi thread test
    sc = RunOnce(con, lambda *args, **kwargs: id(args[0]))
    test = [df1, df2] * 10
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        executor.map(lambda df: assert_or_throw(sc(df, arr) == df, "failed"), test)
    assert 2 == arr[0]



class NoneArgumentError(ValueError):
    """Exception on None argument
    """

    def __init__(self, message: str):
        super().__init__(message)


class InvalidOperationError(Exception):
    """Exception on invalid operations
    """

    def __init__(self, message: str = ""):
        super().__init__(message)



import os
from os.path import exists

from pytest import raises
from triad.collections.fs import FileSystem, _FSPath


def test__FSPath():
    p = _FSPath("/a//b.txt")
    assert "" == p.scheme
    assert "/" == p.root
    assert "a/b.txt" == p.relative_path

    p = _FSPath("//a.txt")
    assert "" == p.scheme
    assert "/" == p.root
    assert "a.txt" == p.relative_path

    p = _FSPath("file://a/b.txt")
    assert "" == p.scheme
    assert "/" == p.root
    assert "a/b.txt" == p.relative_path

    p = _FSPath("hdfs://a/b.txt")
    assert "hdfs" == p.scheme
    assert "hdfs://a" == p.root
    assert "b.txt" == p.relative_path

    p = _FSPath("temp://a/b")
    assert "temp" == p.scheme
    assert "temp://a" == p.root
    assert "b" == p.relative_path

    p = _FSPath("/temp:/a/b")
    assert "temp" == p.scheme
    assert "temp://a" == p.root
    assert "b" == p.relative_path

    raises(ValueError, lambda: _FSPath(None))
    raises(ValueError, lambda: _FSPath(""))
    raises(ValueError, lambda: _FSPath("a.txt"))
    raises(ValueError, lambda: _FSPath("hdfs://"))


def test_fs(tmpdir):
    p1 = os.path.join(tmpdir, "a")
    p2 = os.path.join(tmpdir, "b")
    assert not os.path.exists(p1)
    assert not os.path.exists(p2)
    fs = MockFS()
    fs.makedirs(p1)
    fs.makedirs(p2)
    assert os.path.exists(p1) and os.path.isdir(p1)
    assert os.path.exists(p2) and os.path.isdir(p2)
    assert 1 == fs.create_called
    fs.makedirs("temp://x/y")
    fs.makedirs("temp://y/z")
    assert 3 == fs.create_called
    fs.makedirs("mem://x/y")
    fs.makedirs("mem://y/z")
    assert 5 == fs.create_called
    fs.writetext(os.path.join(p1, "a.txt"), "xyz")
    fs.copy(os.path.join(p1, "a.txt"), "mem://y/z/a.txt")
    assert "xyz" == fs.readtext("mem://y/z/a.txt")
    assert not fs.exists("mem://y/z/w/a.txt")
    assert 5 == fs.create_called
    fs.writetext("mem://from/a.txt", "hello")
    fs.copy("mem://from/a.txt", "mem://to/a.txt")
    assert "hello" == fs.readtext("mem://to/a.txt")
    assert 7 == fs.create_called
    

class MockFS(FileSystem):
    def __init__(self):
        super().__init__()
        self.create_called = 0

    def create_fs(self, root):
        self.create_called += 1
        return super().create_fs(root)
